# -*- coding: utf-8 -*-
"""
Exceptions raised by MTH5


Created on Wed May 13 19:07:21 2020

@author: jpeacock
"""

# Schema Error
class MTSchemaError(Exception):
    pass


class MTTimeError(Exception):
    pass


class MTH5Error(Exception):
    pass


class MTH5TableError(Exception):
    pass


class MTTSError(Exception):
    pass
